struct SSeries
{
    string mName;
    double mVar;
    double mMean;
    double mPoints[];
    double mSyncCloses[];
};

SSeries iMultipliers[], iRotations[], iResults[];

void SetupPCA(const int rows, const int cols)
{
    SetupMatrix(iResults, rows, cols);
    SetupMatrix(iRotations, rows, cols);
    SetupMatrix(iMultipliers, rows, cols);
}

void SetupMatrix(SSeries& series[], const int rows, const int cols)
{
    ArrayResize(series, rows);
    
    for (int k = 0; k < rows; k++)
    {
        ArrayResize(series[k].mPoints, cols);
        ArrayResize(series[k].mSyncCloses, cols);
    }
}

void ClearMatrix(SSeries& series[], const int rows, const int cols)
{
    for (int k = 0; k < rows; k++)
    {
        ArrayFill(series[k].mPoints, 0, cols, 0);
        ArrayFill(series[k].mSyncCloses, 0, cols, 0);
    }
}

void CopyMatrix(SSeries& source[], SSeries& destination[], const int rows, const int cols)
{
    for (int k = 0; k < rows; k++)
    {
        for (int n = 0; n < cols; n++)
        {
            destination[k].mPoints[n] = source[k].mPoints[n];
        }
    }
}

double GetBarByTime(const string symbol, const ENUM_TIMEFRAMES period, const datetime time)
{
    datetime last;
    SeriesInfoInteger(symbol, period, SERIES_LASTBAR_DATE, last);
    return Bars(symbol, period, time, last) - 1;
}

int Synchronize(SSeries& series[], const ENUM_TIMEFRAMES period, const int symbols, const int depth)
{
    int seconds = PeriodSeconds(period);

    for (int k = 0; k < symbols; k++)
    {
        double closes[];
        int copied = CopyClose(series[k].mName, period, 0, depth, closes);
        
        if (copied == depth)
        {
            int closeIndex = depth - 1;
            datetime currentTime = GetLastTime(series, period, symbols);
            series[k].mSyncCloses[closeIndex] = closes[closeIndex];
        
            for (int n = depth - 1; n > 0; n--)
            {
                if (GetBarByTime(series[k].mName, period, currentTime) > 0 && closeIndex > 0)
                {
                    closeIndex--;
                }

                series[k].mSyncCloses[n] = closes[closeIndex];
                currentTime -= seconds;
            }
            
            series[k].mSyncCloses[0] = series[k].mSyncCloses[closeIndex];
        }
    
        if (copied < depth) 
        {
            return(0);
        }
    }
    
    return(1);
}

string GetSymbolNames(const string names, const string base)
{
    string symbols = names;

    if (StringLen(names) < 1)
    {
        int count = SymbolsTotal(false);
        
        for (int k = 0; k < count; k++)
        {
            string symbol = SymbolName(k, false);
            
            if (StringFind(symbol, base) > -1)
            {
                symbols += symbol + (k == count ? "" : ",");
            }
        }
    }

    return(symbols);
}

int ListSymbols(SSeries& series[], const string names)
{
    string symbols[];
    int index = 0, count = StringSplit(names, ',', symbols);

    for (int k = 0; k < count; k++)
    {
        if (SymbolSelect(symbols[k], true) && SymbolInfoInteger(symbols[k], SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL)
        {
            ArrayResize(series, index + 1); 
            series[index++].mName = symbols[k];
        }
    }
    
    return(index);
}

datetime GetLastTime(SSeries& series[], const ENUM_TIMEFRAMES period, const int count)
{
    datetime startTime = 0;

    for (int k = 0; k < count; k++)
    {
        datetime times[];
        CopyTime(series[k].mName, period, 0, 1, times);
        startTime = startTime > times[0] ? startTime : times[0];
    }

    return(startTime);
}

void GetBaseMatrix(SSeries& series[], const int order, const int depth)
{
    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < depth; n++)
        {
            series[k].mPoints[n] = 1000 * MathLog(series[k].mSyncCloses[depth - n - 1]);
        }
    }
}

void GetMeans(SSeries& series[], const int order, const int depth)
{
    for (int k = 0; k < order; k++)
    {
        double mean = 0;
        
        for (int n = 0; n < depth; n++)
        {
            mean += series[k].mPoints[n];
        }
        
        series[k].mMean = mean / depth;

        for (int n = 0; n < depth; n++)
        {
            series[k].mPoints[n] -= series[k].mMean;
        }
    }
}

void GetCovMatrix(SSeries& series[], SSeries& covs[], const int order, const int depth)
{
    GetMeans(series, order, depth);

    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            double variance = 0;

            for (int i = 0; i < depth; i++)
            {
                variance += series[k].mPoints[i] * series[n].mPoints[i];
            }
            
            covs[k].mPoints[n] = variance / depth;
            
            if (k == n)
            {
                covs[k].mVar = covs[k].mPoints[n];
            }
        }
    }
}

void GetCorMatrix(SSeries& series[], SSeries& cors[], const int order)
{
    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            cors[k].mPoints[n] = series[k].mPoints[n] / MathSqrt(series[k].mVar * series[n].mVar);
        }
    }
}

double GetMaxInLuMatrix(SSeries& series[], int& X, int& Y, const int order)
{
    double max = 0;
    
    for (int k = 0; k < order; k++)
    {
        for (int n = k + 1; n < order; n++)
        {
            double candidate = MathAbs(series[k].mPoints[n]);
            
            if (candidate > max)
            {
                max = candidate;
                X = k;
                Y = n;
            }
        }
    }
    
    return max;
}

void MultiplyMatrix(SSeries& series[], SSeries& A[], SSeries& B[], const int order, const bool transposeFirst)
{
    ClearMatrix(iResults, order, order);

    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            iResults[k].mPoints[n] = 0;

            for (int i = 0; i < order; i++) 
            {
                iResults[k].mPoints[n] += transposeFirst ? A[i].mPoints[k] * B[i].mPoints[n] : A[k].mPoints[i] * B[i].mPoints[n];
            }
        }
    }
    
    for (int k = 0; k < order; k++) 
    {
        for (int n = 0; n < order; n++) 
        {
            series[k].mPoints[n] = iResults[k].mPoints[n];
        }
    }
}

double RotateMatrix(SSeries& series[], SSeries& rotations[], int X, int Y, const int order)
{
    double range = series[X].mPoints[X] == series[Y].mPoints[Y] ? 1 : series[X].mPoints[X] - series[Y].mPoints[Y];

    ClearMatrix(iMultipliers, order, order);
    
    double angle = MathArctan(2 * series[X].mPoints[Y] / range) / 2;
    double angleSin = MathSin(angle);
    double angleCos = MathCos(angle);

    for (int k = 0; k < order; k++) 
    {
        ArrayFill(rotations[k].mPoints, 0, order, 0);
        rotations[k].mPoints[k] = iMultipliers[k].mPoints[k] = (k == X || k == Y) ? angleCos : 1;
    }
    
    rotations[Y].mPoints[X] = iMultipliers[Y].mPoints[X] =  angleSin;
    rotations[X].mPoints[Y] = iMultipliers[X].mPoints[Y] = -angleSin;
    
    MultiplyMatrix(series, iMultipliers, series, order, true);
    MultiplyMatrix(series, series, iMultipliers, order, false);
    
    return(1);
}

void GetEigenValues(SSeries& series[], SSeries& eigens[], SSeries& vectors[], const int order, const double precision, const int iterations)
{
    int row = 0, col = 0;

    for (int k = 0; k < order; k++)
    {
        for (int n = 0; n < order; n++)
        {
            eigens[k].mPoints[n] = series[k].mPoints[n];
        }
    }

    ClearMatrix(iRotations, order, order);

    for (int k = 0; k < iterations; k++)
    {
        double max = GetMaxInLuMatrix(eigens, row, col, order);
        
        if (max < precision) 
        {
            break;
        }
        
        if (k > 0)
        {
            RotateMatrix(eigens, iRotations, row, col, order);
            MultiplyMatrix(vectors, vectors, iRotations, order, false);
        }
        else
        {
            RotateMatrix(eigens, vectors, row, col, order);
        }
    }
}

int GetOptimalVector(SSeries& eigens[], const int order, const bool minimize)
{
    int index = 0;
    double issue = minimize ? 1000000 : 0;

    for (int k = 0; k < order; k++) 
    {
        double eigen = MathAbs(eigens[k].mPoints[k]);
        bool condition = minimize ? eigen < issue : eigen > issue;
    
        if (condition)
        {
            issue = eigen;
            index = k;
        }
    }
    
    return(index);
}

int GetIndexValues(SSeries& series[], SSeries& vectors[], const int order, const int depth, const int index)
{
    double summary = 0;

    for (int k = 0; k < order; k++) 
    {
        summary += vectors[k].mPoints[index];
        
        for (int n = 0; n < depth; n++) 
        {
            series[k].mSyncCloses[n] = MathPow(series[k].mSyncCloses[n], vectors[k].mPoints[index]);
        }
    }

    for (int n = 0; n < depth; n++) 
    {
        double product = 1;
    
        for (int k = 0; k < order; k++) 
        {
            product *= MathPow(series[k].mSyncCloses[n], vectors[k].mPoints[index]);
        }
        
        vectors[0].mSyncCloses[n] = MathPow(product, 1.0 / summary);
    }
    
    return(index);
}

void ShowVector(SSeries& series[], SSeries& vectors[], const int order, const int index)
{
    double max = 0;    
    datetime times[];
    int X = 10, Y = 5, size = 8, precision = 5, offset = int(ChartGetDouble(0, CHART_SHIFT_SIZE) * ChartGetInteger(0, CHART_WIDTH_IN_PIXELS) / 100);

    for (int k = 0; k < order; k++) 
    {
        max = MathAbs(vectors[k].mPoints[index]) > max ? MathAbs(vectors[k].mPoints[index]) : max;
    }

    int maxPixel = offset - 155;

    for (int k = 0; k < order; k++) 
    {
        string alignment = (vectors[k].mPoints[index] > 0 ? " " : "") + DoubleToString(vectors[k].mPoints[index], precision);
        string label = StringFormat("%s  | %s", series[k].mName, alignment);
        int level = int(MathAbs(vectors[k].mPoints[index]) * maxPixel / max);

        ShowItem(OBJ_LABEL, series[k].mName + "A", label, X, Y, size, clrBlack);
        ShowItem(OBJ_RECTANGLE_LABEL, series[k].mName + "B", series[k].mName, X + level + 105, Y + 2, size, vectors[k].mPoints[index] > 0 ? clrDeepSkyBlue : clrTomato, level, 10);
        
        Y += size * 2;
    }
}

void ShowItem(const ENUM_OBJECT type, const string name, const string value, const int X, const int Y, const int size, const color clr, const int W = 0, const int H = 0)
{
    long chart = ChartID();
    int window = ChartWindowFind();

    if (ObjectFind(chart, name) < 0)
    {
        ObjectCreate(chart, name, type, window, 0, 0);
    }

    ObjectSetInteger(chart, name, OBJPROP_ANCHOR, ANCHOR_RIGHT_UPPER);
    ObjectSetInteger(chart, name, OBJPROP_CORNER,CORNER_RIGHT_UPPER);
    ObjectSetInteger(chart, name, OBJPROP_FONTSIZE, size);
    ObjectSetInteger(chart, name, OBJPROP_XDISTANCE, X);
    ObjectSetInteger(chart, name, OBJPROP_YDISTANCE, Y);
    ObjectSetInteger(chart, name, OBJPROP_COLOR, clr);
    ObjectSetString(chart, name, OBJPROP_TEXT, value);
    
    if (type == OBJ_RECTANGLE_LABEL) 
    {
        ObjectSetInteger(chart, name, OBJPROP_BGCOLOR, clr);
        ObjectSetInteger(chart, name, OBJPROP_BACK, false);
        ObjectSetInteger(chart, name, OBJPROP_FILL, true);
        ObjectSetInteger(chart, name, OBJPROP_WIDTH, 0);
        ObjectSetInteger(chart, name, OBJPROP_YSIZE, H);
        ObjectSetInteger(chart, name, OBJPROP_XSIZE, W);
    }
}

bool SetLine(const string name, const datetime time)
{
    long chart = ChartID();
    int window = ChartWindowFind();

    if (ObjectFind(chart, name) < 0)
    {
        ObjectCreate(chart, name, OBJ_VLINE, window, 0, 0);
    }
    
    ObjectSetInteger(chart, name, OBJPROP_RAY, true);
    ObjectSetInteger(chart, name, OBJPROP_COLOR, clrRed);
    ObjectSetInteger(chart, name, OBJPROP_STYLE, STYLE_DOT);
    
    ObjectMove(chart, name, 0, time, 0);
    
    return true;
}

void HidePanel()
{
    long chart = ChartID();
    int window = ChartWindowFind();
    ObjectsDeleteAll(chart, window, OBJ_VLINE);
    ObjectsDeleteAll(chart, window, OBJ_LABEL);
    ObjectsDeleteAll(chart, window, OBJ_RECTANGLE_LABEL);
}

void DebugMatrix(SSeries& series[], const bool source)
{
    string result = "\n";
    int order = ArraySize(series);
    
    for (int i = 0; i < order; i++)
    {
       for (int j = 0; j < order; j++)
       {
           result += DoubleToString(source ? series[i].mSyncCloses[j] : series[i].mPoints[j], 5) + " ";
       }
       
       result += "\n";
    }
    
    Comment(result);
}